__title__ = 'ofxstatement plugin for dutch financial institutions'
__package_name__ = 'ofxstatement-dutch'
__author__ = "Gert-Jan Paulissen"
__description__ = ' '.join('''
OFXStatement plugin for dutch financial institutions like DEGIRO, ICSCards,
ING, KNAB and ASN.
'''.strip().split())
__email__ = "gert.jan.paulissen@gmail.com"
__version__ = '1.7.0'
__license__ = "GPLv3"
__copyright__ = 'Copyright 2020-2024 Gert-Jan Paulissen'
__url__ = "https://github.com/gpaulissen/ofxstatement-dutch"

if __name__ == '__main__':
    print(__version__)



# -*- coding: utf-8 -*-
from typing import Iterable, Set, Optional, List, Any, TextIO
import csv
import sys
import datetime
import logging
from decimal import Decimal

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import CsvStatementParser
from ofxstatement.exceptions import ParseError
from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(CsvStatementParser):
    """

    These are the first two lines of a DEGIRO CSV file:

    Datum,Tijd,Valutadatum,Product,ISIN,Omschrijving,FX,Mutatie,,Saldo,,\
Order Id
    30-12-2019,15:58,30-12-2019,MORGAN STANLEY EUR LIQUIDITY FUND,\
LU1959429272,"Conversie geldmarktfonds: Koop 0,000016 @ 9.961,4715 EUR",,,,\
EUR,"13,87",


    These fields are from the Statement class:

    id = ""

    # Date transaction was posted to account (booking date)
    date = datetime.now()

    memo = ""

    # Amount of transaction
    amount = D(0)

    # additional fields
    payee = ""

    # Date user initiated transaction, if known (transaction date)
    date_user = datetime.now()

    # Check (or other reference) number
    check_no = ""

    # Reference number that uniquely identifies the transaction. Can be used in
    # addition to or instead of a check_no
    refnum = ""

    # Transaction type, must be one of TRANSACTION_TYPES
    "CREDIT",       # Generic credit
    "DEBIT",        # Generic debit
    "INT",          # Interest earned or paid
    "DIV",          # Dividend
    "FEE",          # FI fee
    "SRVCHG",       # Service charge
    "DEP",          # Deposit
    "ATM",          # ATM debit or credit
    "POS",          # Point of sale debit or credit
    "XFER",         # Transfer
    "CHECK",        # Check
    "PAYMENT",      # Electronic payment
    "CASH",         # Cash withdrawal
    "DIRECTDEP",    # Direct deposit
    "DIRECTDEBIT",  # Merchant initiated debit
    "REPEATPMT",    # Repeating payment/standing order
    "OTHER"         # Other

    trntype = "CHECK"

    # Optional BankAccount instance
    bank_account_to = None

    """

    date_format = "%d-%m-%Y"

    # 0-based
    mappings = {
        # id (determined later)
        'date': 0,
        'memo': 5,
        'amount': 8,  # Without a header
        # payee
        # date_user
        # check_no
        # refnum
        # trntype (determined later)
        # bank_account_to
    }

    unique_id_set: Set[str]

    def __init__(self, fin: TextIO, account_id: str) -> None:
        # Python 3 needed
        super().__init__(fin)
        # Use the BIC code for ING Netherlands
        self.statement = Statement(bank_id="STDGNL21",
                                   account_id=account_id,
                                   currency="EUR",
                                   # Not yet, just a CHECKING account
                                   # self.statement.account_type = "MONEYMRKT"
                                   account_type="CHECKING")  # My Statement
        self.unique_id_set = set()
        self.header = [["Datum",
                        "Tijd",
                        "Valutadatum",
                        "Product",
                        "ISIN",
                        "Omschrijving",
                        "FX",
                        "Mutatie",
                        "",
                        "Saldo",
                        "",
                        "Order Id"]]

    def parse(self) -> Statement:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """

        # Python 3 needed
        stmt: Statement = super().parse()

        try:
            assert len(self.header) == 0, \
                "Header not completely read: {}".format(str(self.header))
        except Exception as e:
            raise ParseError(0, str(e))

        # GJP 2020-03-03
        # No need to (re)calculate the balance since there is no history.
        # But set the dates.
        stmt.start_balance = stmt.end_balance = None
        if stmt.lines:
            stmt.start_date = min(sl.date for sl in stmt.lines)
            # end date is exclusive for OFX
            stmt.end_date = max(sl.date for sl in stmt.lines)
            stmt.end_date += datetime.timedelta(days=1)

        logger.debug('stmt: %r', stmt)

        return stmt

    def split_records(self) -> Iterable[Any]:
        """Return iterable object consisting of a line per transaction
        """
        return csv.reader(self.fin, delimiter=',')

    def parse_record(self, line: List[str]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        logger.debug('header count: %d; line #%d: %s',
                     len(self.header),
                     self.cur_record,
                     line)

        # First record(s) must be the header
        if len(self.header) >= 1:
            # Remove it since it need not be checked anymore
            hdr = self.header.pop(0)
            logger.debug('header: %s', hdr)
            assert line == hdr, \
                "Expected: {}\ngot: {}".format(hdr, line)
            return None

        # Python 3 needed
        stmt_line: StatementLine = super().parse_record(line)

        # Remove zero-value notifications
        if stmt_line.amount is None or stmt_line.amount == 0:
            return None

        # Forget conversions
        if line[self.mappings['amount'] - 1] != 'EUR':
            return None

        if stmt_line.memo in ['Dividend', 'Dividendbelasting']:
            stmt_line.trntype = "DIV"
        elif stmt_line.memo == 'Rente':
            stmt_line.trntype = "INT"
        elif stmt_line.memo == 'DEGIRO transactiekosten':
            stmt_line.trntype = "FEE"
        elif stmt_line.memo[0:25] == 'DEGIRO Aansluitingskosten':
            stmt_line.trntype = "SRVCHG"
        elif stmt_line.memo == 'Terugstorting':
            stmt_line.trntype = "XFER"
        elif stmt_line.memo in ['Storting', 'iDEAL storting', 'iDEAL Deposit']:
            stmt_line.trntype = "DEP"
        elif stmt_line.amount < 0:  # pragma: no cover
            stmt_line.trntype = "DEBIT"
        else:
            stmt_line.trntype = "CREDIT"

        if stmt_line.trntype not in ["XFER", "DEP"]:
            return None

        # Determine some fields not in the self.mappings
        # A hack but needed to use the adjust method
        stmt_line.__class__ = StatementLine
        stmt_line.adjust(self.unique_id_set)

        # Product known?
        if line[self.mappings['memo'] - 2]:  # pragma: no cover
            stmt_line.memo += ' ' + line[self.mappings['memo'] - 2]
            # ISIN known?
            if line[self.mappings['memo'] - 1]:
                stmt_line.memo +=\
                    ' (' + line[self.mappings['memo'] - 1] + ')'

        return stmt_line

    def parse_decimal(self, value: str) -> Decimal:
        return super().parse_decimal(value) if value else Decimal(0)


class Plugin(BasePlugin):
    """DEGIRO trader platform, The Netherlands, CSV (https://www.degiro.nl/)
    """
    def get_parser(self, f: str) -> Parser:
        fin = open(f, "r", encoding="ISO-8859-1") if isinstance(f, str) else f
        try:
            account_id = self.settings['account_id']
        except Exception:
            raise RuntimeError("""
Please define an 'account_id' in the ofxstatement configuration.

Run

$ ofxstatement edit-config

for more information.
""")
        return Parser(fin, account_id)



# -*- coding: utf-8 -*-
from typing import Set, Optional, List, Iterator, Any, TextIO

import csv
import sys
import datetime
import logging

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import CsvStatementParser
from ofxstatement.exceptions import ParseError, ValidationError
from ofxstatement.statement import BankAccount

from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(CsvStatementParser):
    """

    These are the first three lines of a KNAB CSV file:

    KNAB EXPORT;;;;;;;;;;;;;;;;
    Rekeningnummer;Transactiedatum;Valutacode;CreditDebet;Bedrag;\
Tegenrekeningnummer;Tegenrekeninghouder;Valutadatum;Betaalwijze;\
Omschrijving;Type betaling;Machtigingsnummer;Incassant ID;Adres;Referentie;\
Boekdatum;
    "NL99KNAB9999999999";"21-06-2019";"EUR";"D";"827,75";\
"FR9999999999999999999999999";"Gert Janssen (FR)";"21-06-2019";"Overboeking";\
"Tresorie";"";"";"";"";"XXXXXXXXXXXXXXXX";\
"21-06-2019";

    These fields are from the Statement class:

    id = ""

    # Date transaction was posted to account (booking date)
    date = datetime.now()

    memo = ""

    # Amount of transaction
    amount = D(0)

    # additional fields
    payee = ""

    # Date user initiated transaction, if known (transaction date)
    date_user = datetime.now()

    # Check (or other reference) number
    check_no = ""

    # Reference number that uniquely identifies the transaction. Can be used in
    # addition to or instead of a check_no
    refnum = ""

    # Transaction type, must be one of TRANSACTION_TYPES
    "CREDIT",       # Generic credit
    "DEBIT",        # Generic debit
    "INT",          # Interest earned or paid
    "DIV",          # Dividend
    "FEE",          # FI fee
    "SRVCHG",       # Service charge
    "DEP",          # Deposit
    "ATM",          # ATM debit or credit
    "POS",          # Point of sale debit or credit
    "XFER",         # Transfer
    "CHECK",        # Check
    "PAYMENT",      # Electronic payment
    "CASH",         # Cash withdrawal
    "DIRECTDEP",    # Direct deposit
    "DIRECTDEBIT",  # Merchant initiated debit
    "REPEATPMT",    # Repeating payment/standing order
    "OTHER"         # Other

    trntype = "CHECK"

    # Optional BankAccount instance
    bank_account_to = None

    """

    date_format = "%d-%m-%Y"

    # 0-based
    mappings = {
        # id (determined later)
        'date': 15,  # Boekdatum
        'memo': 9,  # Omschrijving
        'amount': 4,  # Bedrag
        'payee': 6,  # Tegenrekeninghouder
        'date_user': 1,  # Transactiedatum
        # check_no
        'refnum': 14,  # Referentie
        # trntype (determined later)
        'bank_account_to': 5,  # Tegenrekeningnummer
    }

    unique_id_set: Set[str]

    # Other mappings not used by parser.CsvStatementParser
    ACCOUNT = 0  # Rekeningnummer
    CD = 3  # CreditDebit

    def __init__(self, fin: TextIO) -> None:
        # Python 3 needed
        super().__init__(fin)
        # Use the BIC code for KNAB Online, The Netherlands
        self.statement = Statement(bank_id="KNABNL2H",
                                   account_id=None,
                                   currency="EUR")  # My Statement
        self.unique_id_set = set()
        self.header = [['KNAB EXPORT'],
                       ['Rekeningnummer',
                        'Transactiedatum',
                        'Valutacode',
                        'CreditDebet',
                        'Bedrag',
                        'Tegenrekeningnummer',
                        'Tegenrekeninghouder',
                        'Valutadatum',
                        'Betaalwijze',
                        'Omschrijving',
                        'Type betaling',
                        'Machtigingsnummer',
                        'Incassant ID',
                        'Adres',
                        'Referentie',
                        'Boekdatum']]

    def parse(self) -> StatementLine:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """

        # Python 3 needed
        stmt: StatementLine = super().parse()

        try:
            assert len(self.header) == 0, \
                "Header not completely read: {}".format(str(self.header))
        except Exception as e:
            raise ParseError(0, str(e))

        try:
            assert len(stmt.lines) > 0, "No statement lines read"

            # GJP 2020-03-03
            # No need to (re)calculate the balance since there is no history.
            # But set the dates.
            stmt.start_balance = stmt.end_balance = None
            stmt.start_date = min(sl.date for sl in stmt.lines)
            # end date is exclusive for OFX
            stmt.end_date = max(sl.date for sl in stmt.lines)
            stmt.end_date += datetime.timedelta(days=1)
        except Exception as e:
            raise ValidationError(str(e), stmt)

        return stmt

    def split_records(self) -> Iterator[Any]:
        """Return iterable object consisting of a line per transaction
        """
        return csv.reader(self.fin, delimiter=';')

    def parse_record(self, line: List[str]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        try:
            logger.debug('header count: %d; line #%d: %s',
                         len(self.header),
                         self.cur_record,
                         line)

            # First record(s) must be the header
            if len(self.header) >= 1:
                # Remove it since it need not be checked anymore
                hdr = self.header.pop(0)
                line = list(filter(None, line))
                logger.debug('header: %s', hdr)
                assert line == hdr, \
                    "Expected: {}\ngot: {}".format(hdr, line)
                return None

            # line[self.ACCOUNT] contains the account number
            if self.statement.account_id:
                assert self.statement.account_id == \
                    line[self.ACCOUNT], \
                    "Only one account is allowed; previous account: {}, \
this line's account: {}".format(self.statement.account_id,
                                line[self.ACCOUNT])
            else:
                self.statement.account_id = line[self.ACCOUNT]

            assert line[self.CD] in ['D', 'C'], \
                "Element {} is not D/C in line {}".format(self.CD, str(line))

            if line[self.CD] == 'D':
                line[self.mappings['amount']] =\
                    '-' + line[self.mappings['amount']]

            if line[self.mappings['bank_account_to']]:
                line[self.mappings['payee']] =\
                    "{} ({})".format(line[self.mappings['payee']],
                                     line[self.mappings['bank_account_to']])

            # Python 3 needed
            stmt_line: StatementLine = super().parse_record(line)

            # Remove zero-value notifications
            if stmt_line.amount == 0:
                return None

            # Determine some fields not in the self.mappings
            # A hack but needed to use the adjust method
            stmt_line.__class__ = StatementLine
            stmt_line.adjust(self.unique_id_set)

            if stmt_line.amount < 0:
                stmt_line.trntype = "DEBIT"
            else:
                stmt_line.trntype = "CREDIT"

            if stmt_line.bank_account_to:
                stmt_line.bank_account_to = \
                    BankAccount(bank_id='',
                                acct_id=stmt_line.bank_account_to)
        except Exception as e:
            raise ParseError(self.cur_record, str(e))

        return stmt_line


class Plugin(BasePlugin):
    """KNAB Online Bank, The Netherlands, CSV (https://www.knab.nl/)
    """
    def get_parser(self, f: str) -> Parser:
        fin = open(f, "r", encoding="ISO-8859-1") if isinstance(f, str) else f
        return Parser(fin)



import os
import pytest
from unittest import TestCase
from decimal import Decimal
from datetime import datetime

from ofxstatement.statement import StatementLine
from ofxstatement.exceptions import ParseError
from ofxstatement.plugins.nl.degiro import Plugin


class ParserTest(TestCase):

    def test_ok(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'Account_20190101_20200317.csv')
        parser = Plugin(None, {'account_id': 'ABC'}).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "STDGNL21")
        self.assertEqual(statement.account_id, "ABC")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertIsNone(statement.start_balance)
        self.assertIsNone(statement.end_balance)
        self.assertEqual(statement.start_date,
                         datetime.strptime("19-06-2019",
                                           parser.date_format))
        self.assertEqual(statement.end_date,
                         datetime.strptime("22-06-2019",
                                           parser.date_format))

        lines = []

        # ignore lines without money

        lines.append(StatementLine(date='30-12-2019',
                                   memo='Conversie geldmarktfonds: \
Koop 0,000016 @ 9.961,4715 EUR MORGAN STANLEY EUR LIQUIDITY FUND \
(LU1959429272)',
                                   amount=''))
        lines.append(StatementLine(date='27-12-2019',
                                   memo='Dividend VANECK AEX (NL0009272749)',
                                   amount='0,19'))
        lines.append(StatementLine(date='27-12-2019',
                                   memo='Dividendbelasting VANECK AEX \
(NL0009272749)',
                                   amount='-0,03'))
        lines.append(StatementLine(date='26-09-2019',
                                   memo='Conversie geldmarktfonds: \
Koop 0,000039 @ 9.975,9152 EUR MORGAN STANLEY EUR LIQUIDITY FUND \
(LU1959429272)',
                                   amount=''))
        lines.append(StatementLine(date='25-09-2019',
                                   memo='Dividend VANECK AEX (NL0009272749)',
                                   amount='0,46'))
        lines.append(StatementLine(date='25-09-2019',
                                   memo='Dividendbelasting VANECK AEX \
(NL0009272749)',
                                   amount='-0,07'))
        lines.append(StatementLine(date='27-06-2019',
                                   memo='Conversie geldmarktfonds: \
Koop 0,001337 @ 9.986,9062 EUR MORGAN STANLEY EUR LIQUIDITY FUND \
(LU1959429272)',
                                   amount=''))
        lines.append(StatementLine(date='26-06-2019',
                                   memo='Dividend VANECK AEX (NL0009272749)',
                                   amount='1,06'))
        lines.append(StatementLine(date='26-06-2019',
                                   memo='Dividendbelasting VANECK AEX \
(NL0009272749)',
                                   amount='-0,16'))
        lines.append(StatementLine(date='26-06-2019',
                                   memo='Dividend VANECK ESG EW \
(NL0010408704)',
                                   amount='4,40'))
        lines.append(StatementLine(date='26-06-2019',
                                   memo='Dividendbelasting VANECK ESG EW \
(NL0010408704)',
                                   amount='-0,66'))
        lines.append(StatementLine(date='26-06-2019',
                                   memo='Dividend ISHARES AEX (IE00B0M62Y33)',
                                   amount='8,72'))
        lines.append(StatementLine(date='21-06-2019',
                                   memo='Terugstorting',
                                   amount='-334,35'))
        lines.append(StatementLine(date='21-06-2019',
                                   memo='Terugstorting #2',
                                   amount='-334,35'))
        lines.append(StatementLine(date='21-06-2019',
                                   memo='DEGIRO transactiekosten VANECK ESG EW\
 (NL0010408704)',
                                   amount='-2,00'))
        lines.append(StatementLine(date='21-06-2019',
                                   memo='DEGIRO Aansluitingskosten',
                                   amount='-0,13'))
        lines.append(StatementLine(date='21-06-2019',
                                   memo='Verkoop 4 @ 84,12 EUR VANECK ESG EW\
 (NL0010408704)',
                                   amount='336,48'))
        lines.append(StatementLine(date='19-06-2019',
                                   memo='iDEAL storting',
                                   amount='557,10'))
        lines.append(StatementLine(date='19-06-2019',
                                   memo='Verkoop 10 @ 55,92 EUR ISHARES AEX\
 (IE00B0M62Y33)',
                                   amount='559,20'))
        lines.append(StatementLine(date='04-06-2019',
                                   memo='Rente',
                                   amount='-0,01'))
        lines.append(StatementLine(date='03-05-2019',
                                   memo='Rente',
                                   amount='-0,01'))
        lines.append(StatementLine(date='01-04-2019',
                                   memo='Rente',
                                   amount='-0,02'))
        lines.append(StatementLine(date='27-03-2019',
                                   memo='Dividend VANECK AEX (NL0009272749)',
                                   amount='0,25'))
        lines.append(StatementLine(date='27-03-2019',
                                   memo='Dividendbelasting VANECK AEX \
(NL0009272749)',
                                   amount='-0,04'))
        lines.append(StatementLine(date='27-03-2019',
                                   memo='Dividend VANECK ESG EW\
 (NL0010408704)',
                                   amount='1,00'))
        lines.append(StatementLine(date='27-03-2019',
                                   memo='Dividendbelasting VANECK ESG EW \
(NL0010408704)',
                                   amount='-0,15'))
        lines.append(StatementLine(date='27-03-2019',
                                   memo='Dividend ISHARES AEX (IE00B0M62Y33)',
                                   amount='1,97'))
        # in USD
        # lines.append(StatementLine(date='04-03-2019',
        #                            memo='Rente',
        #                            amount='-0,02'))
        lines.append(StatementLine(date='01-02-2019',
                                   memo='Rente',
                                   amount='-0,02'))
        lines.append(StatementLine(date='02-01-2019',
                                   memo='Rente #2',
                                   amount='-0,02'))

        lines = [line
                 for line in lines
                 if line.amount and line.memo in ['iDEAL storting',
                                                  'Terugstorting',
                                                  'Terugstorting #2']]
        self.assertEqual(len(statement.lines), len(lines))

        for idx, line in enumerate(statement.lines):
            self.assertEqual(line.date,
                             datetime.strptime(lines[idx].date,
                                               parser.date_format))
            self.assertEqual(line.memo, lines[idx].memo)
            self.assertEqual(line.amount,
                             Decimal(lines[idx].amount.replace(",", ".").
                                     replace(" ", "")))

    @pytest.mark.xfail(raises=RuntimeError)
    def test_no_config(self):
        """No attribute
        """
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'Account_20190101_20200317.csv')
        Plugin(None, None).get_parser(text_filename)

    @pytest.mark.xfail(raises=RuntimeError)
    def test_no_config_account_id(self):
        """No attribute
        """
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'Account_20190101_20200317.csv')
        Plugin(None, {'ACCOUNT_ID'}).get_parser(text_filename)

    @pytest.mark.xfail(raises=ParseError)
    def test_no_header(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'empty.csv')
        parser = Plugin(None, {'account_id': 'ABC'}).get_parser(text_filename)

        # And parse csv:
        parser.parse()



import io
import os
from textwrap import dedent
from unittest import TestCase
from decimal import Decimal
import pytest
from datetime import datetime

from ofxstatement.exceptions import ParseError, ValidationError

from ofxstatement.plugins.nl.knab import Parser, Plugin


class ParserTest(TestCase):

    def test_basic(self):
        # Lets define some sample csv to parse and write it to file-like object
        csv = dedent('''
KNAB EXPORT;;;;;;;;;;;;;;;;
Rekeningnummer;Transactiedatum;Valutacode;CreditDebet;Bedrag;Tegenrekeningnummer;Tegenrekeninghouder;Valutadatum;Betaalwijze;Omschrijving;Type betaling;Machtigingsnummer;Incassant ID;Adres;Referentie;Boekdatum;
"NL99KNAB9999999999";"26-03-2020";"EUR";"D";"7,02";"NL99ASNB9999999999";"JANSSEN G";"27-03-2020";"Overboeking";"Omschrijving 1";"";"";"";"";"C0C27IP2NC00000A";"28-03-2020";
"NL99KNAB9999999999";"26-03-2020";"EUR";"D";"0,00";"NL99ASNB9999999999";"JANSSEN G";"27-03-2020";"Overboeking";"";"";"";"";"";"C0C27IP2NC00000A";"28-03-2020";
"NL99KNAB9999999999";"27-03-2020";"EUR";"C";"5,00";"50022270";"Gert Janssen";"28-03-2020";"Ontvangen betaling";"Omschrijving 2";"";"";"";"";"C0C27PGFM28ERA34";"29-03-2020";
            ''')
        f = io.StringIO(csv)

        # Create and configure csv parser:
        parser = Parser(f)

        # And parse csv:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "KNABNL2H")
        self.assertEqual(statement.account_id, "NL99KNAB9999999999")
        self.assertEqual(statement.account_type, "CHECKING")

        self.assertIsNone(statement.start_balance)
        self.assertEqual(statement.start_date, datetime.strptime("28-03-2020", parser.date_format))

        self.assertIsNone(statement.end_balance)
        self.assertEqual(statement.end_date, datetime.strptime("30-03-2020", parser.date_format))

        # Amount of 0 is skipped
        self.assertEqual(len(statement.lines), 2)
        self.assertEqual(statement.lines[0].date, datetime.strptime("28-03-2020", parser.date_format))
        self.assertEqual(statement.lines[0].date_user, datetime.strptime("26-03-2020", parser.date_format))
        self.assertEqual(statement.lines[0].amount, Decimal('-7.02'))
        self.assertEqual(statement.lines[0].payee, "JANSSEN G (NL99ASNB9999999999)")
        self.assertEqual(statement.lines[0].memo, "Omschrijving 1")
        self.assertEqual(statement.lines[0].refnum, "C0C27IP2NC00000A")

        self.assertEqual(statement.lines[1].date, datetime.strptime("29-03-2020", parser.date_format))
        self.assertEqual(statement.lines[1].date_user, datetime.strptime("27-03-2020", parser.date_format))
        self.assertEqual(statement.lines[1].amount, Decimal('5.00'))
        self.assertEqual(statement.lines[1].payee, "Gert Janssen (50022270)")
        self.assertEqual(statement.lines[1].memo, "Omschrijving 2")
        self.assertEqual(statement.lines[1].refnum, "C0C27PGFM28ERA34")

    def test_ok(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'Knab_transactieoverzicht_ok.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()
        self.assertEqual(len(statement.lines), 28)
        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "KNABNL2H")
        self.assertEqual(statement.account_id, "NL99KNAB9999999999")
        self.assertEqual(statement.account_type, "CHECKING")

        self.assertIsNone(statement.start_balance)
        self.assertEqual(statement.start_date, datetime.strptime("07-01-2019", parser.date_format))

        self.assertIsNone(statement.end_balance)
        self.assertEqual(statement.end_date, datetime.strptime("22-06-2019", parser.date_format))

        self.assertEqual(sum(sl.amount for sl in statement.lines), Decimal('7.01'))

    @pytest.mark.xfail(raises=ParseError)
    def test_no_header1(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'empty.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        parser.parse()

    @pytest.mark.xfail(raises=ParseError)
    def test_no_header2(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'Knab_transactieoverzicht_no_header2.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        parser.parse()

    @pytest.mark.xfail(raises=ValidationError)
    def test_no_lines(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'Knab_transactieoverzicht_no_lines.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        parser.parse()

    @pytest.mark.xfail(raises=ParseError)
    def test_wrong_account_no(self):
        # Lets define some sample csv to parse and write it to file-like object
        csv = dedent('''
KNAB EXPORT;;;;;;;;;;;;;;;;
Rekeningnummer;Transactiedatum;Valutacode;CreditDebet;Bedrag;Tegenrekeningnummer;Tegenrekeninghouder;Valutadatum;Betaalwijze;Omschrijving;Type betaling;Machtigingsnummer;Incassant ID;Adres;Referentie;Boekdatum;
"NL99KNAB9999999999";"26-03-2020";"EUR";"D";"7,02";"NL99ASNB9999999999";"JANSSEN G";"27-03-2020";"Overboeking";"Omschrijving 1";"";"";"";"";"C0C27IP2NC00000A";"28-03-2020";
"NL00KNAB0000000000";"27-03-2020";"EUR";"C";"5,00";"50022270";"Gert Janssen";"28-03-2020";"Ontvangen betaling";"Omschrijving 2";"";"";"";"";"C0C27PGFM28ERA34";"29-03-2020";
            ''')
        f = io.StringIO(csv)

        # Create and configure csv parser:
        parser = Parser(f)

        # And parse csv:
        parser.parse()

    @pytest.mark.xfail(raises=ParseError)
    def test_no_payee(self):
        # Lets define some sample csv to parse and write it to file-like object
        csv = dedent('''
KNAB EXPORT;;;;;;;;;;;;;;;;
Rekeningnummer;Transactiedatum;Valutacode;CreditDebet;Bedrag;Tegenrekeningnummer;Tegenrekeninghouder;Valutadatum;Betaalwijze;Omschrijving;Type betaling;Machtigingsnummer;Incassant ID;Adres;Referentie;Boekdatum;
"NL99KNAB9999999999";"26-03-2020";"EUR";"D";"7,02";"NL99ASNB9999999999";"JANSSEN G";"27-03-2020";"Overboeking";"Omschrijving 1";"";"";"";"";"C0C27IP2NC00000A";"28-03-2020";
"NL99KNAB9999999999";"27-03-2020";"EUR";"C";"5,00";"";"Gert Janssen";"28-03-2020";"Ontvangen betaling";"Omschrijving 2";"";"";"";"";"C0C27PGFM28ERA34";"29-03-2020";
            ''')
        f = io.StringIO(csv)

        # Create and configure csv parser:
        parser = Parser(f)

        # And parse csv:
        parser.parse()



import os
from unittest import TestCase
from decimal import Decimal
import pytest
from datetime import datetime

from ofxstatement.exceptions import ParseError

from ofxstatement.plugins.nl.ing import Plugin


class ParserTest(TestCase):

    def check(self, parser):
        # And parse csv:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "INGBNL2A")
        self.assertEqual(statement.account_id, "NL99INGB9999999999")
        self.assertEqual(statement.account_type, "CHECKING")

        self.assertIsNone(statement.start_balance)
        self.assertEqual(statement.start_date,
                         datetime.strptime("20191213",
                                           parser.date_format))

        self.assertIsNone(statement.end_balance)
        self.assertEqual(statement.end_date,
                         datetime.strptime("20200214",
                                           parser.date_format))

        # Amount of 0 is skipped
        self.assertEqual(len(statement.lines), 5)
        self.assertEqual(statement.lines[0].amount, Decimal('-1.25'))
        self.assertFalse(statement.lines[0].payee)
        # "Naam / Omschrijving" is prepended to "Mededelingen"
        self.assertEqual(statement.lines[0].memo,
                         "Kosten OranjePakket met korting, \
1 jan t/m 31 jan 2020 ING BANK N.V. Valutadatum: 13-02-2020")

        self.assertEqual(statement.lines[1].amount, Decimal('1.25'))
        self.assertFalse(statement.lines[1].payee)
        # "Naam / Omschrijving" is prepended to "Mededelingen"
        self.assertEqual(statement.lines[1].memo,
                         "Kwijtschelding, Valutadatum: 13-02-2020")

        self.assertEqual(statement.lines[2].amount, Decimal('20.00'))
        # "Naam / Omschrijving" is prepended to "Tegenrekening"
        self.assertEqual(statement.lines[2].payee,
                         "PAULISSEN G J L M (NL99ASNB9999999999)")
        # "Naam / Omschrijving" is NOT prepended to "Mededelingen"
        self.assertEqual(statement.lines[2].memo,
                         "Naam: PAULISSEN G J L M Omschrijving: \
Kosten rekening IBAN: NL99ASNB9999999999 Valutadatum: 13-12-2019")

        self.assertEqual(statement.lines[3].amount, Decimal('-0.31'))
        self.assertFalse(statement.lines[3].payee)
        # "Naam / Omschrijving" is prepended to "Mededelingen"
        self.assertEqual(statement.lines[3].memo,
                         "Kosten OranjePakket, \
25 nov t/m 30 nov 2019 ING BANK N.V. Valutadatum: 13-12-2019")

        self.assertEqual(statement.lines[4].amount, Decimal('-0.31'))
        self.assertFalse(statement.lines[4].payee)
        # "Naam / Omschrijving" is prepended to "Mededelingen"
        self.assertEqual(statement.lines[4].memo,
                         "Kosten OranjePakket, \
25 nov t/m 30 nov 2019 ING BANK N.V. Valutadatum: 13-12-2019 #2")

    def test_ok(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'ing_ok.csv')
        self.check(Plugin(None, None).get_parser(text_filename))

    def test_ok_Mutatiesoort(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'ing_ok_Mutatiesoort.csv')
        self.check(Plugin(None, None).get_parser(text_filename))

    def test_ok_Mutatiesoort_Extra(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'ing_ok_Mutatiesoort_Extra.csv')
        self.check(Plugin(None, None).get_parser(text_filename))

    def test_ok_Mutatiesoort_Extra_Unquoted(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'ing_ok_Mutatiesoort_Extra_Unquoted.csv')
        self.check(Plugin(None, None).get_parser(text_filename))

    @pytest.mark.xfail(raises=ParseError)
    def test_fail(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'ing_fail.csv')
        parser = Plugin(None, None).get_parser(text_filename)
        # Lets define some sample csv to parse and write it to file-like object

        # And parse csv:
        parser.parse()

    @pytest.mark.xfail(raises=ParseError)
    def test_no_header(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'empty.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        parser.parse()

    def test_balance(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'NL99INGB9999999999_25-11-2019_30-05-2020.csv')
        parser = Plugin(None, None).get_parser(text_filename)
        # Lets define some sample csv to parse and write it to file-like object

        # And parse csv:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "INGBNL2A")
        self.assertEqual(statement.account_id, "NL99INGB9999999999")
        self.assertEqual(statement.account_type, "CHECKING")

        self.assertIsNone(statement.start_balance)
        self.assertIsNone(statement.start_date)

        self.assertEqual(statement.end_balance, Decimal('13.20'))
        self.assertEqual(statement.end_date,
                         datetime.strptime("2020-05-31",  # plus 1 day
                                           parser.date_format))

        self.assertEqual(len(statement.lines), 0)



# -*- coding: utf-8 -*-
from typing import Optional, List, Iterator, Any, Dict, TextIO

import re
import csv
import sys
import datetime
import logging
from decimal import Decimal

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import CsvStatementParser
from ofxstatement.exceptions import ParseError
from ofxstatement.statement import BankAccount

from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(CsvStatementParser):
    """

    Nr  Veldnaam                     Formaat        Omschrijving                                                                            Voorbeeld
    --  --------                     -------        ------------                                                                            ---------
     0  Boekingsdatum                dd­mm­jjjj     Dit veld geeft de datum weer waarop de transactie daadwerkelijk heeft plaatsgevonden.   3­4­2000
     1  Opdrachtgeversrekening       X (18)         Uw ASN­Rekening (IBAN).                                                                 NL01ASNB0123456789
     2  Tegenrekeningnummer          X (34)         Dit veld bevat het rekeningnummer (IBAN) naar of waarvan de transactie afkomstig is.    NL01BANK0123456789
                                                    Het IBAN telt maximaal 34 alfanumerieke tekens en heeft een vaste lengte per land
                                                    Het IBAN bestaat uit een landcode (twee letters), een controlegetal (twee cijfers)
                                                    en een (voor bepaalde landen aangevuld) nationaal rekeningnummer.
     3  Naam tegenrekening           X (70)         Hier wordt de naam van de tegenrekening vermeld.                                        jansen
                                                    De naam is maximaal 70 posities lang en wordt in kleine letters weergegeven.
     4  Adres                                       Dit veld is niet in gebruik.
     5  Postcode                                    Dit veld is niet in gebruik.
     6  Plaats                                      Dit veld is niet in gebruik
     7  Valutasoort rekening         XXX            Dit veld geeft de ISO valutasoort van de rekening weer.                                 EUR
                                                    Een bestand kan verschillende valutasoorten bevatten.
     8  Saldo rekening voor mutatie  ­999999999.99  Geeft het saldo weer van de rekening voordat de mutatie is verwerkt.                    122800.83 of ­123.30
                                                    Als decimaal scheidingsteken wordt een punt gebruikt. Er wordt geen duizend separator
                                                    gebruikt. In het geval van een negatieve waarde wordt het bedrag voorafgegaan van een
                                                    - (min) teken.
     9  Valutasoort mutatie          XXX            Dit veld geeft de ISO valutasoort van de mutatie weer.                                  EUR
                                                    Een bestand kan verschillende valutasoorten bevatten.
    10  Transactiebedrag             ­999999999.99  Geeft het transactiebedrag weer. Als decimaal scheidingsteken wordt een punt gebruikt.  238.45 of ­43.90
                                                    Een negatief bedrag wordt voorafgegaan door een - (min) teken.
    11  Journaaldatum                dd­mm­jjjj     De journaaldatum is de datum waarop een transactie in de systemen van ASN Bank wordt    21­01­2000
                                                    geboekt. Dit hoeft niet noodzakelijkerwijs gelijk te zijn aan de boekingsdatum.
    12  Valutadatum                  dd­mm­jjjj     Dit veld geeft de valutadatum weer. De valutadatum is de datum waarop een bedrag        01­04­2001
                                                    rentedragend wordt.
    13  Interne transactiecode       9999           Dit is een interne transactiecode zoals die door de ASN Bank wordt gebruikt. Deze       8810 of 9820
                                                    transactiecodes kunnen gebruikt worden om heel verfijnd betaalde transacties te
                                                    herkennen. Zoals een bijboeking van een geldautomaat opname. Er kan geen garantie
                                                    worden gegeven dat deze codes in de toekomst hetzelfde blijven en/of dat er codes
                                                    vervallen en/of toegevoegd zullen worden.
    14  Globale transactiecode       XXX            De globale transactiecode is een vertaling van de interne transactiecode. Gebruikte     GEA of BEA of VV
                                                    afkortingen zijn bijvoorbeeld BEA voor een betaalautomaat opname of GEA voor een
                                                    geldautomaat opname. In de bijlage wordt een overzicht gegeven van alle gebruikte
                                                    afkortingen.
                                                    Zie ook Bijlage 1: Gebruikte boekingscodes
    15  Volgnummer transactie        N (8)          Geeft het transactievolgnummer van de transactie weer. Dit volgnummer vormt samen met   90043054
                                                    de journaaldatum een uniek transactie id.
    16  Betalingskenmerk             X (16)         Het betalingskenmerk bevat de meest relevante gegevens zoals die door de betaler zijn   ’factuur 9234820’
                                                    opgegeven. Zoals debiteuren nummer en/of factuurnummer. Het betalingskenmerk wordt
                                                    tussen enkele quotes (’) geplaatst.
    17  Omschrijving                 X (140)        De omschrijving zoals die bij de overboeking is opgegeven. De omschrijving kan          ’02438000140032extra trekking werelddierendag 4info’
                                                    maximaal 140 posities beslaan.
    18  Afschriftnummer              N (3)          Het nummer van het afschrift waar de betreffende boeking op staat vermeld.              42

    ===

    Bijlage 1: Gebruikte boekingscodes

    ACC Acceptgirobetaling AF Afboeking
    AFB Afbetalen
    BEA Betaalautomaat BIJ Bijboeking
    BTL Buitenlandse Overboeking
    CHP Chipknip
    CHQ Cheque
    COR Correctie
    DIV Diversen
    EFF Effectenboeking
    ETC Euro traveller cheques GBK GiroBetaalkaart
    GEA Geldautomaat
    INC Incasso
    IDB iDEAL betaling
    IMB iDEAL betaling via mobiel IOB Interne Overboeking
    KAS Kas post
    KTN Kosten/provisies
    KST Kosten/provisies
    OVB Overboeking
    PRM Premies
    PRV Provisies
    RNT Rente
    STO Storno
    TEL Telefonische Overboeking VV Vreemde valuta

    ===

    These fields are from the Statement class:

    id = ""

    # Date transaction was posted to account (booking date)
    date = datetime.now()

    memo = ""

    # Amount of transaction
    amount = D(0)

    # additional fields
    payee = ""

    # Date user initiated transaction, if known (transaction date)
    date_user = datetime.now()

    # Check (or other reference) number
    check_no = ""

    # Reference number that uniquely identifies the transaction. Can be used in
    # addition to or instead of a check_no
    refnum = ""

    # Transaction type, must be one of TRANSACTION_TYPES
    "CREDIT",       # Generic credit
    "DEBIT",        # Generic debit
    "INT",          # Interest earned or paid
    "DIV",          # Dividend
    "FEE",          # FI fee
    "SRVCHG",       # Service charge
    "DEP",          # Deposit
    "ATM",          # ATM debit or credit
    "POS",          # Point of sale debit or credit
    "XFER",         # Transfer
    "CHECK",        # Check
    "PAYMENT",      # Electronic payment
    "CASH",         # Cash withdrawal
    "DIRECTDEP",    # Direct deposit
    "DIRECTDEBIT",  # Merchant initiated debit
    "REPEATPMT",    # Repeating payment/standing order
    "OTHER"         # Other

    trntype = "CHECK"

    # Optional BankAccount instance
    bank_account_to = None

    """

    date_format: str = "%d-%m-%Y"

    # 0-based
    mappings: Dict[str, int] = {
        # id (determined later)
        'date': 11,
        'memo': 17,
        'amount': 10,
        'payee': 3,  # if bank_account_to is filled
        'date_user': 0,
        # check_no
        # refnum
        # trntype (determined later)
        'bank_account_to': 2,
    }

    def __init__(self,
                 fin: TextIO,
                 account_id: Optional[str] = None) -> None:
        # Python 3 needed
        super().__init__(fin)
        # Use the BIC code for ASN Bank
        self.statement = Statement(bank_id="ASNBNL21",
                                   account_id=account_id,
                                   currency="EUR")  # My Statement

    def parse(self) -> Statement:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """

        # Python 3 needed
        stmt: Statement = super().parse()

        # GJP 2020-03-03
        # No need to (re)calculate the balance since there is no history.
        # But set the dates.
        if stmt.lines:
            stmt.start_date = min(sl.date for sl in stmt.lines)
            # end date is exclusive for OFX
            stmt.end_date = max(sl.date for sl in stmt.lines)
            stmt.end_date += datetime.timedelta(days=1)
            stmt.start_balance = Decimal(stmt.lines[0].start_balance)
            stmt.end_balance = Decimal(stmt.lines[-1].start_balance) + stmt.lines[-1].amount

        return stmt

    def split_records(self) -> Iterator[Any]:
        """Return iterable object consisting of a line per transaction
        """
        # strip quotes around memo
        return csv.reader(self.fin, delimiter=',', quotechar="'")

    def parse_record(self,
                     line: List[str]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        stmt_line: StatementLine

        try:
            logger.debug('line #%d: %s',
                         self.cur_record,
                         line)

            stmt_line = self.parse_transaction(line)

        except Exception as e:
            raise ParseError(self.cur_record, str(e))

        return stmt_line

    def parse_transaction(self,
                          line: List[str]) -> Optional[StatementLine]:
        start_balance: int = 8
        transaction_nr: int = 15

        # line[1] contains the account number
        if self.statement.account_id:
            assert self.statement.account_id == line[1], \
                "Only one account is allowed; previous account: {}, \
this line's account: {}".format(self.statement.account_id, line[1])
        else:
            self.statement.account_id = line[1]

        if line[self.mappings['bank_account_to']]:
            line[self.mappings['payee']] =\
                "{} ({})".format(line[self.mappings['payee']],
                                 line[self.mappings['bank_account_to']])
        else:
            line[self.mappings['payee']] = ''

        # Python 3 needed
        stmt_line: StatementLine = super().parse_record(line)

        # Remove zero-value notifications
        if stmt_line.amount == 0:
            return None

        stmt_line.__class__ = StatementLine

        # The unique id is a combination of 'Journaaldatum' and 'Volgnummer transactie'
        # Let id be <Journaaldatum in yyyymmdd format>.<Volgnummer transactie>
        assert self.mappings['date'] == 11  # Journaaldatum
        assert transaction_nr == 15  # Volgnummer transactie

        assert line[self.mappings['date']]
        dd_mm_yyyy: str = str(line[self.mappings['date']])
        stmt_line.id = "{}{}{}.{}".format(dd_mm_yyyy[6:],
                                          dd_mm_yyyy[3:5],
                                          dd_mm_yyyy[0:2],
                                          line[transaction_nr])

        stmt_line.start_balance = Decimal(str(line[start_balance])) if line[start_balance] is not None else Decimal(0)

        if stmt_line.amount < 0:
            stmt_line.trntype = "DEBIT"
        else:
            stmt_line.trntype = "CREDIT"

        if stmt_line.bank_account_to:
            stmt_line.bank_account_to = \
                BankAccount(bank_id='',
                            acct_id=stmt_line.bank_account_to)
        else:
            stmt_line.bank_account_to = None

        return stmt_line


class Plugin(BasePlugin):
    """ASN Bank, The Netherlands, CSV (https://www.asnbank.nl/)
    """
    def get_parser(self, filename: str) -> Parser:
        p = re.compile('transactie-historie_(NL\\d+ASNB\\d+)_\\d+\\.csv')
        m = p.search(filename)
        account_id: Optional[str] = None
        if m:
            account_id = m.group(1)
        fin = open(filename, "r")  # , encoding="ISO-8859-1")
        return Parser(fin, account_id)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""Setup
"""

import sys
from setuptools import find_packages
from setuptools.command.test import test as TestCommand
from distutils.core import setup

# To prevent importing about and thereby breaking the coverage info we use this
# exec hack
about = {}
with open('__about__.py') as fp:
    exec(fp.read(), about)


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to pytest')]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ''

    def run_tests(self):
        import shlex
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


tests_require = [
    'pyyaml',
    'pytest',
    'pytest-cache',
    'pytest-cover',
    'pytest-flakes',
    'pytest-pycodestyle',
    'flake8',
]


if sys.argv[-1] == 'info':
    for k, v in about.items():
        print('%s: %s' % (k, v))
    sys.exit()


if __name__ == '__main__':
    with open('README.md') as f:
        readme = f.read()

    with open('CHANGELOG.md') as f:
        changes = f.read()

    setup(
        name=about['__package_name__'],
        version=about['__version__'],
        author=about['__author__'],
        author_email=about['__email__'],
        description=about['__description__'],
        url=about['__url__'],
        license=about['__license__'],
        keywords=["ofx", "banking", "statement", "beancount",
                  "degiro", "ing", "icscards", "knab"],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        namespace_packages=["ofxstatement", "ofxstatement.plugins"],
        long_description=readme + changes,
        long_description_content_type='text/markdown',
        include_package_data=True,
        install_requires=['ofxstatement>0.6.4'],
        tests_require=tests_require,
        setup_requires=[
            'setuptools>=39.1.0',
        ],
        zip_safe=True,
        cmdclass={'test': PyTest},
        extras_require={'test': tests_require},
        classifiers=[
            'Development Status :: 6 - Mature',
            'Programming Language :: Python :: 3',
            'Natural Language :: English',
            'Topic :: Office/Business :: Financial :: Accounting',
            'Topic :: Utilities',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)'
        ],
        entry_points={
            'ofxstatement':
            ['nl-degiro = ofxstatement.plugins.nl.degiro:Plugin',
             'nl-icscards = ofxstatement.plugins.nl.icscards:Plugin',
             'nl-ing = ofxstatement.plugins.nl.ing:Plugin',
             'nl-knab = ofxstatement.plugins.nl.knab:Plugin',
             'nl-asn = ofxstatement.plugins.nl.asn:Plugin']
        },
    )



# -*- coding: utf-8 -*-
import os
from unittest import TestCase
from decimal import Decimal
import pytest
from datetime import datetime

from ofxstatement.plugins.nl.icscards import Plugin


class ParserTest(TestCase):

    def test_missing_column(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'icscards.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-1311.73'))
        self.assertEqual(statement.start_date,
                         datetime.strptime("2019-08-21",
                                           parser.date_format))
        self.assertEqual(statement.end_balance, Decimal('-1320.55'))
        self.assertEqual(statement.end_date,
                         datetime.strptime("2019-09-17",
                                           parser.date_format))

        self.assertEqual(len(statement.lines), 25)
        self.assertEqual(statement.lines[0].amount, Decimal('1311.73'))
        self.assertEqual(statement.lines[1].date,
                         datetime.strptime("2019-08-21",
                                           parser.date_format))
        self.assertEqual(statement.lines[1].amount, Decimal('-7.99'))
        self.assertEqual(statement.lines[12].payee, "HOTEL MERCURE")
        self.assertEqual(statement.lines[12].memo, "MONTIGNY LE B (FR)")
        self.assertEqual(statement.lines[13].payee,
                         "NEWREST WAGONS LITS FRANC")
        self.assertEqual(statement.lines[13].memo, "PARIS (FR)")
        self.assertEqual(statement.lines[14].payee, "SNCF")
        self.assertEqual(statement.lines[14].memo, "PARIS 8 (FR)")
        self.assertEqual(statement.lines[24].amount, Decimal('-6.15'))

    def test_big(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'icscards_big.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-893.31'))
        self.assertEqual(statement.start_date,
                         datetime.strptime("2018-12-21",
                                           parser.date_format))
        self.assertEqual(statement.end_balance, Decimal('-1156.34'))
        self.assertEqual(statement.end_date,
                         datetime.strptime("2019-01-17", parser.date_format))

    def test_error(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'icscards_error.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-1753.69'))
        self.assertEqual(statement.end_balance, Decimal('-1325.11'))

        self.assertEqual(statement.lines[6].payee, "THY 2357312380512")
        self.assertEqual(statement.lines[6].memo, "Istanbul (US)")
        self.assertEqual(statement.lines[16].payee, "TOTAL 4375462")
        self.assertEqual(statement.lines[16].memo, "33PESSAC (FR)")

    def test_equal_transactions(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'icscards_equal_transactions.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-1311.73'))
        self.assertEqual(statement.end_balance, Decimal('-1320.55'))

        self.assertEqual(statement.lines[7].amount, Decimal('-5.05'))
        self.assertEqual(statement.lines[7].memo, 'MONTIGNY LE B (FR)')
        self.assertEqual(statement.lines[8].amount, Decimal('-5.05'))
        self.assertEqual(statement.lines[8].memo, 'MONTIGNY LE B (FR) #2')
        self.assertNotEqual(statement.lines[7].id, statement.lines[8].id)

    def test_2025_03(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'icscards-2025-03.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-0.99'))
        self.assertEqual(statement.end_balance, Decimal('-1.98'))

        self.assertEqual(statement.lines[0].amount, Decimal('-0.99'))
        self.assertEqual(statement.lines[0].memo, 'ITUNES.COM (IE)')

    def test_2025_04(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'icscards-2025-04.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-1.98'))
        self.assertEqual(statement.end_balance, Decimal('-2.97'))

        self.assertEqual(statement.lines[0].amount, Decimal('-0.99'))
        self.assertEqual(statement.lines[0].memo, 'ITUNES.COM (IE)')

    def test_no_balance(self):
        # Create and configure parser:
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here,
                                     'samples',
                                     'icscards_no_balance.txt')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ABNANL2A")
        self.assertEqual(statement.account_id, "99999999999")
        self.assertEqual(statement.account_type, "CHECKING")
        self.assertEqual(statement.start_balance, Decimal('-13.08'))
        self.assertEqual(statement.end_balance, Decimal('0.00'))

        self.assertEqual(statement.lines[0].amount, Decimal('13.08'))
        self.assertEqual(statement.lines[0].memo, 'IDEAL BETALING, DANK U')

    @pytest.mark.xfail(raises=AttributeError)
    def test_fail(self):
        """'Parser' object has no attribute 'bank_id'
        """
        here = os.path.dirname(__file__)
        pdf_filename = os.path.join(here, 'samples', 'blank.pdf')
        parser = Plugin(None, None).get_parser(pdf_filename)

        # And parse:
        parser.parse()
        if parser.bank_id:
            pass



# -*- coding: utf-8 -*-
import re
from typing import Set, Union

from ofxstatement.statement import StatementLine as BaseStatementLine
from ofxstatement.statement import Statement as BaseStatement
from ofxstatement.exceptions import ValidationError
from ofxstatement.statement import generate_unique_transaction_id
from datetime import datetime, date
import logging


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def _to_date(d_t: Union[date, datetime]) -> date:
    return d_t.date() if isinstance(d_t, datetime) else d_t


class Statement(BaseStatement):
    def assert_valid(self) -> None:

        logger.debug("self: type: %s; contents: %s", type(self), self)
        try:
            super().assert_valid()
            dates = [sl.date for sl in self.lines if sl.date is not None]
            # An ING CSV may be a balance file resulting in 0 lines
            if len(dates) == 0:
                return
            assert self.start_date, "The statement start date should be set"
            assert self.end_date, "The statement end date should be set"
            # check self.start_date
            min_date = _to_date(min(dates))
            start_date = _to_date(self.start_date)
            assert start_date and min_date and start_date <= min_date, \
                "The statement start date ({}) should at most be the smallest \
statement line date ({})".format(start_date, min_date)
            # check self.end_date
            max_date = _to_date(max(dates))
            end_date = _to_date(self.end_date)
            assert end_date and max_date and end_date > max_date, \
                "The statement end date ({}) should be greater than the \
largest statement line date ({})".format(end_date, max_date)
        except Exception as e:
            raise ValidationError(str(e), self)


class StatementLine(BaseStatementLine):
    """Statement line data with an adjust method.
    """
    def adjust(self, unique_id_set: Set[str]) -> None:
        if self.id:
            return

        self.id = \
            generate_unique_transaction_id(self, unique_id_set)
        m = re.match(r'([0-9a-f]+)(-\d+)?$', self.id)
        assert m, "Id should match hexadecimal digits, \
optionally followed by a minus and a counter: '{}'".format(self.id)
        if m.group(2):
            counter = int(m.group(2)[1:])
            # include counter so the memo gets unique
            self.memo = self.memo + ' #' + str(counter + 1)  # type: ignore



import os
from unittest import TestCase
from decimal import Decimal
import pytest
from datetime import datetime

from ofxstatement.exceptions import ParseError

from ofxstatement.plugins.nl.asn import Plugin


class ParserTest(TestCase):

    def test_ok(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'transactie-historie_NL00ASNB9999999999_20220717204133.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        statement = parser.parse()

        self.assertEqual(statement.currency, 'EUR')
        self.assertEqual(statement.bank_id, "ASNBNL21")
        self.assertEqual(statement.account_id, "NL00ASNB9999999999")
        self.assertEqual(statement.account_type, "CHECKING")

        self.assertEqual(statement.start_balance, Decimal('130.44'))
        self.assertEqual(statement.start_date,
                         datetime.strptime("17-06-2022",
                                           parser.date_format))

        self.assertEqual(statement.end_balance, Decimal('644.24') + Decimal('-560.00'))
        self.assertEqual(statement.end_date,
                         datetime.strptime("12-07-2022",  # +1
                                           parser.date_format))

        self.assertEqual(len(statement.lines), 11)
        self.assertEqual(statement.lines[0].id, '20220617.51392971')
        self.assertEqual(statement.lines[0].amount, Decimal('223.77'))
        self.assertEqual(statement.lines[0].bank_account_to.acct_id, 'NL99ASNB0000000000')
        self.assertEqual(statement.lines[0].payee, 'XXXXXXXXX Z Z Z Z (NL99ASNB0000000000)')
        self.assertEqual(statement.lines[0].date,
                         datetime.strptime("17-06-2022",
                                           parser.date_format))
        self.assertEqual(statement.lines[0].date_user,
                         statement.lines[0].date)

        self.assertEqual(statement.lines[2].id, '20220625.50951652')
        self.assertIsNone(statement.lines[2].bank_account_to)
        self.assertFalse(statement.lines[2].payee)
        self.assertEqual(statement.lines[2].memo,
                         "Kosten gebruik betaalrekening inclusief 1 betaalpas")

        self.assertEqual(statement.lines[5].id, '20220629.50139616')
        self.assertEqual(statement.lines[5].date,
                         datetime.strptime("29-06-2022",
                                           parser.date_format))
        self.assertEqual(statement.lines[5].date_user,
                         datetime.strptime("28-06-2022",
                                           parser.date_format))

    @pytest.mark.xfail(raises=ParseError)
    def test_fail(self):
        here = os.path.dirname(__file__)
        text_filename = os.path.join(here, 'samples', 'transactie-historie_fail.csv')
        parser = Plugin(None, None).get_parser(text_filename)

        # And parse csv:
        parser.parse()



# -*- coding: utf-8 -*-
from typing import Iterable, Set, Optional, List, Iterator, Any, Union

import sys
import locale
import re
import io
from decimal import Decimal
from datetime import datetime
from subprocess import check_output, CalledProcessError
import logging

from ofxstatement.plugin import Plugin as BasePlugin
from ofxstatement.parser import StatementParser as BaseStatementParser

from ofxstatement.plugins.nl.statement import Statement, StatementLine

# Need Python 3 for super() syntax
assert sys.version_info[0] >= 3, "At least Python 3 is required."

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class Parser(BaseStatementParser):  # type: ignore
    unique_id_set: Set[str]

    def __init__(self, fin: Iterable[str]) -> None:
        super().__init__()
        self.statement = Statement(bank_id=None,
                                   account_id=None,
                                   currency='EUR')  # My Statement
        self.fin = fin
        self.unique_id_set = set()

    def parse(self) -> Statement:
        """Main entry point for parsers

        super() implementation will call to split_records and parse_record to
        process the file.
        """
        stmt: Optional[Statement] = None
        # Save locale
        current_locale = locale.setlocale(category=locale.LC_ALL)
        # Need to parse "05 mei" i.e. "05 may"
        locale.setlocale(category=locale.LC_ALL, locale="nl_NL")
        try:
            # Python 3 needed
            stmt = super().parse()

            if stmt and stmt.lines:
                stmt.start_date = min(sl.date for sl in stmt.lines if sl.date)

            stmt.assert_valid()
        finally:
            locale.setlocale(category=locale.LC_ALL, locale=current_locale)

        return stmt

    @staticmethod
    def get_amount(amount_in: str, transaction_type_in: str) -> Decimal:
        sign_out: int = 1
        amount_out: Union[str, Decimal]

        # determine sign_out
        assert isinstance(transaction_type_in, str)
        assert transaction_type_in in ['Af', 'Bij', '  ']

        if transaction_type_in == 'Af':
            sign_out = -1

        # determine amount_out
        assert isinstance(amount_in, str)
        # Amount something like 1.827,97, € 1.827,97 (both dutch) or 1,827.97?
        # Since April 2025 it may be €1.827,97 as well
        m = re.search(r'^(\S+\s|\D)?([0-9,.]+)$', amount_in)
        assert m is not None
        amount_out = m.group(2)
        if amount_out[-3] == ',':
            amount_out = amount_out.replace('.', '').replace(',', '.')

        # convert to str to keep just the last two decimals
        amount_out = sign_out * Decimal(str(amount_out))
        logger.debug("get_amount(%s, %s) = %s", amount_in, transaction_type_in, amount_out)
        return amount_out

    def split_records(self) -> Iterator[Any]:
        """Return iterable object consisting of a line per transaction
        """
        def convert_str_to_list(str: str,
                                max_items: Optional[int] = None,
                                sep: str = r'\s\s+|\t|\n') -> List[str]:
            return [x for x in re.split(sep, str)[0:max_items]]

        first_line = True
        first_line_row = ['International Card Services BV', 'www.icscards.nl']

        new_page = False
        new_page_row = ['Datum', 'ICS-klantnummer', 'Volgnummer', 'Bladnummer']

        balance = False
        balance_row = ['Vorig openstaand saldo', 'Totaal ontvangen betalingen',
                       'Totaal nieuwe uitgaven', 'Nieuw openstaand saldo']

        # 01 sep          01 sep            IDEAL BETALING, DANK U                                                                                                          1.311,73       Bij
        # 21 feb         22 feb            APPLE.COM/BILL                                  ITUNES.COM                       IE                                                  0,99   Af
        # Since April 2025 (period after month):
        # 21 mrt.        22 mrt.           APPLE.COM/BILL                                  ITUNES.COM                       IE                                                  0,99   Af
        statement_expr = \
            re.compile(r'^\d\d [a-z]{3}\.?\s+\d\d [a-z]{3}\.?.+[0-9,.]+\s+(Af|Bij|  )$')
        country = re.compile("^[A-Z][A-Z]$")

        for line in self.fin:
            line = line.strip()

            logger.debug('line: %s', line)

            # to ease the parsing pain
            row: List[str] = convert_str_to_list(line)

            logger.debug('row: %s', row)
            logger.debug("before parsing line: self.statement: %s; first_line: %s; new_page: %s; balance: %s",
                         self.statement,
                         first_line,
                         new_page,
                         balance)

            if first_line and len(row) > 1:
                assert row == first_line_row, \
                    "Expected: {0}\nActual: {1}".format(first_line_row, row)
                first_line = False

            if len(row) == 2 and row[1][0:5] == 'BIC: ':
                self.statement.bank_id = row[1][5:]

            elif row == new_page_row:
                new_page = True
            elif new_page:
                new_page = False
                # exclusive in ICSCards
                self.statement.end_date = \
                    datetime.strptime(row[0],
                                      '%d %B %Y')
                self.statement.account_id = row[1]

            elif row == balance_row:
                balance = True
            elif balance:
                balance = False
                # Row has 4 times an amount and after each amount there may be "Af" or "Bij"
                assert len(row) >= 4 and len(row) <= 8
                if row[-1] not in ['Af', 'Bij']:
                    row.append('  ')
                if row[-3] not in ['Af', 'Bij']:
                    row.insert(-2, '  ')
                if row[-5] not in ['Af', 'Bij']:
                    row.insert(-4, '  ')
                if row[-7] not in ['Af', 'Bij']:
                    row.insert(-6, '  ')
                assert len(row) == 8
                for i in range(int(len(row) / 2)):
                    assert row[i * 2] not in ['Af', 'Bij', '  ']
                    assert row[i * 2 + 1] in ['Af', 'Bij', '  ']

                self.statement.start_balance = Parser.get_amount(row[0],
                                                                 row[1])
                self.statement.end_balance = Parser.get_amount(row[-2],
                                                               row[-1])

            elif re.search(statement_expr, line):
                # payee, place and country may be something like:
                #
                # THY|2357312380512|Istanbul|US
                #
                # hence four columns instead of three, so combine the first two
                for i in reversed(range(len(row))):
                    if country.match(row[i]):
                        # Should have 4 columns to the left. If not: reduce.
                        while i > 4:
                            row[2] += ' ' + row[3]
                            del row[3]
                            i -= 1
                        break

                if len(row) >= 6 and len(row) <= 7 and len(row[2]) > 25:
                    # 04 sep | 05 sep | NEWREST WAGONS LITS FRANCPARIS ...
                    # =>
                    # 04 sep | 05 sep | NEWREST WAGONS LITS FRANC ...

                    row.insert(2, row[2][0:25])
                    row[3] = row[3][25:]
                logger.debug('yield row: %s', row)
                yield row
            logger.debug("after parsing line : self.statement: %s; first_line: %s; new_page: %s; balance: %s",
                         self.statement,
                         first_line,
                         new_page,
                         balance)

    def parse_record(self, row: List[str]) -> Optional[StatementLine]:
        """Parse given transaction line and return StatementLine object
        """

        def add_years(dt: datetime, years: int) -> datetime:
            """Return a date that's `years` years after the date (or datetime)
            object `d`. Return the same calendar date (month and day) in the
            destination year, if it exists, otherwise use the following day
            (thus changing February 29 to March 1).

            """
            result: datetime = dt.replace(year=dt.year + years, month=3, day=1) \
                if dt.month == 2 and dt.day == 29 \
                else dt.replace(year=dt.year + years)
            logger.debug("add_years(%s, %d) = %s", dt, years, result)
            return result

        def get_date(d_m: str) -> Optional[datetime]:
            # Without a year it will be 1900 so add the year
            assert self.statement.end_date and self.statement.end_date.year
            d_m_y = "{} {}".format(d_m, self.statement.end_date.year)
            dt: Optional[datetime] = datetime.strptime(d_m_y, '%d %b %Y')
            # But now the resulting date may be more than the end date
            # (d_m in december and end date in january)
            if dt and dt > self.statement.end_date:
                dt = add_years(dt, -1)
            assert dt is None or dt <= self.statement.end_date
            return dt

        logger.debug('parse_record(%s)', str(row))
        assert len(row) in [5, 7, 8]

        stmt_line: Optional[StatementLine] = None
        # GJP 2020-03-01
        # Skip transaction date (index 0) since it gives a wrong balance.
        # Use booking date (index 1) in order to get a correct balance.
        # Since April 2025 the month will end in a period ('.').
        date = get_date(row[1][:-1] if row[1].endswith('.') else row[1])

        payee = None
        memo = None
        if len(row) >= 7:
            # payee (2), place (3) and country (4)
            payee = row[2]
            memo = "{0} ({1})".format(row[3], row[4])
        else:
            # description (2)
            memo = row[2]

        # Skip amount in foreign currency
        amount = Parser.get_amount(row[-2], row[-1])

        # Remove zero-value notifications
        if amount != 0:
            stmt_line = StatementLine(date=date,
                                      memo=memo,
                                      amount=amount)
            stmt_line.payee = payee
            stmt_line.adjust(self.unique_id_set)

        logger.debug('stmt_line: %s', stmt_line)
        return stmt_line


class Plugin(BasePlugin):
    """ICSCards, The Netherlands, PDF (https://icscards.nl/)
    """

    def get_file_object_parser(self, fh: Iterable[str]) -> Parser:
        return Parser(fh)

    def get_parser(self, filename: str) -> Parser:
        pdftotext = ["pdftotext", "-layout", filename, '-']
        fh: Iterable[str]

        # Is it a PDF or an already converted file?
        try:
            fh = io.StringIO(check_output(pdftotext).decode())
            # No exception: apparently it is a PDF.
        except CalledProcessError:
            fh = open(filename, "r")

        return self.get_file_object_parser(fh)
